#include "stdafx.h"
#include "LeoHelpers_Scope.h"
#include "LeoHelpers_Window.h"
#include "LeoHelpers_String.h"
#include "LeoHelpers_Misc.h"

// The result may not be reliable for windows owned by other threads. (e.g. If they change between asking for the length and asking for the string.)
BOOL LeoHelpers::GetWindowTextString(HWND hwnd, std::wstring *pStrResult, bool bUseTimeoutEvenIfSameProcess)
{
	pStrResult->clear();

	int iLength = 0;
	
	if (bUseTimeoutEvenIfSameProcess)
	{
		DWORD dwProcessId = 0;
		::GetWindowThreadProcessId(hwnd, &dwProcessId);

		if (dwProcessId != ::GetCurrentProcessId())
		{
			// The normal API has timeouts built-in when going cross-process,
			// and I don't want to change the rules about top-level vs child windows in other processes.
			bUseTimeoutEvenIfSameProcess = false;
		}
	}

	if (bUseTimeoutEvenIfSameProcess)
	{
#ifndef SMTO_ERRORONEXIT
		const UINT SMTO_ERRORONEXIT = 0x0020;
#endif

		DWORD_PTR lr = 0;
		if (!::SendMessageTimeout(hwnd, WM_GETTEXTLENGTH, 0, 0, SMTO_NORMAL|SMTO_ABORTIFHUNG|(LeoHelpers::IsWindowsVistaOrAbove()?SMTO_ERRORONEXIT:0), 100, &lr)
		||	lr > INT_MAX)
		{
			return FALSE;
		}
		iLength = static_cast<int>(lr);
	}
	else
	{
		iLength = ::GetWindowTextLength(hwnd);
	}

	if (0 == iLength)
	{
		return TRUE;
	}

	iLength++; // Not clear if GetWindowTextLength includes the null but GetWindowText does.

	// Unless the buffer size is really big (for a string) already, double it so that
	// if the string has changed to a longer one we still have a good chance of getting it.
	if (iLength < 4096)
	{
		iLength *= 2;
	}

	LeoHelpers::WCharArrayScoper buf(new wchar_t[ iLength ]);

	if (bUseTimeoutEvenIfSameProcess)
	{
#ifndef SMTO_ERRORONEXIT
		const UINT SMTO_ERRORONEXIT = 0x0020;
#endif

		DWORD_PTR lr = 0;
		if (!::SendMessageTimeout(hwnd, WM_GETTEXT, iLength, reinterpret_cast< LPARAM >( buf.Get() ), SMTO_NORMAL|SMTO_ABORTIFHUNG|(LeoHelpers::IsWindowsVistaOrAbove()?SMTO_ERRORONEXIT:0), 100, &lr)
		||	0 == lr)
		{
			return FALSE;
		}
	}
	else
	{
		if (0 == ::GetWindowText(hwnd, buf.Get(), iLength))
		{
			return FALSE;
		}
	}

	*pStrResult = buf.Get();

	return TRUE;
}

LeoHelpers::WindowInfo::WindowInfo(HWND hwnd)
{
	if (!::IsWindow(hwnd))
	{
		LeoHelpers::StringFormatToString(&m_strInfo, L"Invalid window handle %p", hwnd);
	}
	else
	{
		bool bVisible = ::IsWindowVisible(hwnd) ? true : false;

		wchar_t szClass[1024];

		if (0 == ::GetClassName(hwnd, szClass, _countof(szClass)))
		{
			szClass[0] = L'\0';
		}

		std::wstring strTitle;

		if (!LeoHelpers::GetWindowTextString(hwnd, &strTitle, true))
		{
			strTitle = L"<error or timeout>";
		}

		DWORD dwProcess = 0;
		DWORD dwThread = ::GetWindowThreadProcessId(hwnd, &dwProcess);

		bool bIsDesktop = (hwnd == ::GetDesktopWindow());

		LeoHelpers::StringFormatToString(&m_strInfo, L"HWND %p%s, ProcId %ld, ThreadId %ld, %s visible, Class \"%s\", Title \"%s\"", hwnd, bIsDesktop?L" (the desktop)":L"", dwProcess, dwThread, bVisible?L"Is":L"Not", szClass, strTitle.c_str());
	}
}

// Get top-level parent window
HWND LeoHelpers::GetTopLevelWindow(HWND hWnd)
{
	// Loop until we have a top-level window
	while (hWnd != NULL)
	{
		LONG_PTR styles = ::GetWindowLongPtr(hWnd, GWL_STYLE);
		
		if (!(styles&WS_CHILD))
		{
			break; // Either error or top-level window.
		}

		hWnd = ::GetParent(hWnd);
	}

	return hWnd;
}

// Give hwndThis=NULL to test the other window against the current thread.
bool LeoHelpers::SameWindowThread(HWND hwndThis, HWND hwndOther)
{
	DWORD dwThreadIdThis = 0;

	if (hwndThis == NULL)
	{
		dwThreadIdThis = ::GetCurrentThreadId();
	}
	else
	{
		dwThreadIdThis = ::GetWindowThreadProcessId(hwndThis, NULL);
	}

	if (dwThreadIdThis == 0)
	{
		return false;
	}

	// Thread IDs are system-wide so we don't need to check the process ID as well.
	DWORD dwThreadIdOther = ::GetWindowThreadProcessId(hwndOther, NULL);

	return (dwThreadIdThis == dwThreadIdOther);
}

// Check if a window is hung. Can help avoid deadlocks when using multiple UI threads/processes.
// Beware that this won't detect all deadlock situations: The window in question could have a child
// or parent window running on another UI thread which may still deadlock with the calling thread
// when something is done to hwnd which causes it to try and send a message to the other UI thread.
// bBlock maps to SendMessageTimeout's SMTO_BLOCK and prevents the calling thread from processing
// messages sent to it from other threads while waiting for the window in question to respond.
// bAssert set true will fire an assert in debug builds if the window isn't responding.
bool LeoHelpers::IsWindowResponding(HWND hwnd, bool bBlock, bool bAssert)
{
#ifndef SMTO_ERRORONEXIT
	const UINT SMTO_ERRORONEXIT = 0x0020;
#endif

	DWORD_PTR dwpResult = 0; // irrelevant but required.

	// Leo 04/Nov/2010: Removed SMTO_NOTIMEOUTIFNOTHUNG as it effectively makes the timeout about 30 seconds longer, which is not wanted.
	LRESULT lResWindowResponding = ::SendMessageTimeout(hwnd, WM_NULL, 0, 0,
		SMTO_ABORTIFHUNG|(LeoHelpers::IsWindowsVistaOrAbove()?SMTO_ERRORONEXIT:0)|(bBlock?SMTO_BLOCK:SMTO_NORMAL),
		200, &dwpResult);

	if (bAssert)
	{
		assert(lResWindowResponding != 0);
	}

	return (lResWindowResponding != 0);
}
